import OUtils.Kafka

import confluent_kafka
import uuid
import time
import re


class Heartbeat:

    @staticmethod
    def are_topics_skewed():
        # [1] https://docs.confluent.io/current/clients/confluent-kafka-python/#confluent_kafka.admin.PartitionMetadata
        configs = OUtils.Kafka.get_check_configurations("heartbeat")

        heartbeat_topics = OUtils.Kafka.get_topics_metadata(configs["topic_pattern"])

        # Ensure that something is being checked
        assert len(heartbeat_topics) > 0, \
            "No metadata loaded/found for provided topic_pattern '%s'" % configs["topic_pattern"]

        # Ensure quantity of brokers and heartbeat topics matches
        qty_of_heartbeat_topics = len(heartbeat_topics)
        qty_of_brokers = len(OUtils.Kafka.get_brokers_metadata())
        assert qty_of_heartbeat_topics >= qty_of_brokers, \
            "Quantity of heartbeat topics (%d) and brokers (%d) differ" % (qty_of_heartbeat_topics, qty_of_brokers)

        # Ensure the heartbeat topics are not partitioned and not replicated
        for heartbeat_topic in heartbeat_topics:

            assert len(heartbeat_topic.partitions) == 1, \
                "Heartbeat topic '%s' has multiple partitions" % heartbeat_topic

            assert len(heartbeat_topic.partitions[0].replicas) == 1, \
                "Heartbeat topic '%s' is replicated" % heartbeat_topic

        # If all assertions succeeded
        return False

    @staticmethod
    def is_topic_on_homonym_node(topic):
        heartbeat_topics = OUtils.Kafka.get_topics_metadata(topic)

        return heartbeat_topics[0].topic.endswith(str(heartbeat_topics[0].partitions[0].leader))

    @staticmethod
    def run_on_topic(topic):
        # Ensure heartbeat topics are at the right places
        assert OUtils.Kafka.checks.Heartbeat.are_topics_skewed() is False, \
            "Heartbeat topics not ready for check execution (skewed situation)"

        assert OUtils.Kafka.checks.Heartbeat.is_topic_on_homonym_node(topic) is True, \
            "Heartbeat topic '%s' is not hosted on the homonym broker" % topic

        configs = OUtils.Kafka.get_check_configurations("heartbeat")

        # Ensure heartbeat is executed on topics dedicated to heartbeats
        assert re.match(configs["topic_pattern"], topic), \
            "The topic specified '%s' does not match the required pattern '%s'" % (topic, configs["topic_pattern"])

        # Actors
        consumer = Heartbeat.__get_consumer(topic, configs["timeout_assignment"])
        producer = Heartbeat.__get_producer(topic)

        # Generate token
        injected_token = str(uuid.uuid4())

        time_0 = time.clock()

        # Produce a message
        producer.produce(topic=topic, value=injected_token.encode("utf-8"))

        # Consume messages
        messages = consumer.consume(num_messages=1, timeout=configs["timeout_consumption"])

        time_1 = time.clock()

        # Close consumer handler to release partition assignments more quickly (and help the next check execution)
        consumer.close()

        # Check 1: quantity of messages
        assert len(messages) == 1, \
            "Unexpected quantity of messages received (%d)" % len(messages)

        # Check 2: verify content
        loaded_token = str(messages[0].value().decode("utf-8"))
        assert loaded_token == injected_token, \
            "The consumed message (%s) is not consistent with injected token (%s)" % (loaded_token, injected_token)

        # If all good, return heartbeat elapsed time
        return time_1 - time_0

    @staticmethod
    def __get_consumer(topic, timeout):
        properties = OUtils.Kafka.get_base_properties()

        properties["group.id"] = properties["client.id"] = topic  # Use topic name as group.id, so to avoid offset commit overlaps

        consumer = confluent_kafka.Consumer(properties)
        consumer.subscribe([topic])

        for attempt in range(timeout):
            consumer.poll(1)
            if len(consumer.assignment()) > 0:
                break

        assignments = consumer.assignment()

        assert len(assignments) == 1, \
            "Partitions assignment (%d) inconsistent after timeout." % len(assignments)

        # Move offsets to the latest point for each partition
        # assignments[0].offset = confluent_kafka.OFFSET_END

        # consumer.assign(assignments)

        return consumer

    @staticmethod
    def __get_producer(topic):
        properties = OUtils.Kafka.get_base_properties()

        properties["client.id"] = topic

        return confluent_kafka.Producer(properties)



import OUtils.configurations
import OUtils.Kafka.metrics as metrics
import OUtils.Kafka.checks as checks

import confluent_kafka.admin
import re


def get_nodes(port=False):
    result = []
    nodes = OUtils.configurations.get(module="kafka", key="nodes")

    for node in nodes:
        if port is False:
            result.append(node["host"])

        elif port not in node["ports"]:
            raise KeyError("Port '%s' not supported" % port)

        else:
            result.append("%s:%d" % (node["host"], node["ports"][port]))

    assert len(result) >= 3, "Too few nodes (%s) loaded. Please check configuration file." % len(result)

    return result


def get_configuration_files():
    result = OUtils.configurations.get(module="kafka", key="configuration_files")

    assert len(result) > 0, "Too few (%d) configuration files loaded. Please check configurations file." % len(result)

    return result


def get_check_configurations(check_name):
    result = OUtils.configurations.get(module="kafka", key="checks")

    assert check_name in result.keys(), "No configurations found for check (%s)." % check_name

    return result[check_name]


def get_base_properties():
    nodes = OUtils.Kafka.get_nodes(port="kafka")

    properties = OUtils.configurations.get(module="kafka", key="base_properties")
    properties = {**properties, **{"bootstrap.servers": ",".join(nodes)}}

    return properties


def __get_admin_client():
    properties = OUtils.Kafka.get_base_properties()

    return confluent_kafka.admin.AdminClient(properties)


def get_topics_metadata(topic_name_pattern=".*"):
    return [topic_metadata for topic_metadata in __get_admin_client().list_topics().topics.values()
            if re.match(topic_name_pattern, topic_metadata.topic)]


def get_brokers_metadata():
    return __get_admin_client().list_topics().brokers



import os
import sys
import json
import jsonschema

__loaded_configurations = None


def __load_configurations():
    global __loaded_configurations

    if __loaded_configurations is not None:
        return

    configurations_file_path = os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])), "configurations.json")

    if not os.path.exists(configurations_file_path):
        raise FileExistsError("Configuration file '%s' not found" % configurations_file_path)

    with open(configurations_file_path) as json_file:
        __loaded_configurations = json.load(json_file)


def __validate_configurations():
    global __loaded_configurations

    if __loaded_configurations is None:
        raise ImportError("Can't validate configurations (Not loaded yet")

    schema_file_path = os.path.dirname(os.path.realpath(__file__)) + '/configurationSchema.json'

    with open(schema_file_path) as json_file:
        schema = json.load(json_file)

        jsonschema.validate(instance=__loaded_configurations, schema=schema)


def get(module, key):
    global __loaded_configurations

    if __loaded_configurations is None:
        __load_configurations()
        __validate_configurations()

    if module not in __loaded_configurations:
        raise KeyError("Configuration module '%s' not found" % module)

    if key not in __loaded_configurations[module]:
        raise KeyError("Configuration '%s' not found under module '%s'" % (key, module))

    return __loaded_configurations[module][key]



import OUtils.Kafka

import subprocess
import time


class RollingRestart:

    def __init__(self):
        pass

    def __verify_cluster_clean_state(self):
        assert OUtils.Kafka.checks.are_configuration_files_aligned() is True, "Configuration files not aligned"
        assert OUtils.Kafka.metrics.get_total_under_replicated_replicas() == 0, "There are under-replicated replicas"

        return True

    def __get_confirmation(self, string):
        confirmation = input(string)

        if confirmation.lower() in ["yes", "y"]:
            return True

        return False

    def __execute_kafka_service_restart(self, node):
        subprocess.run("bash -c \"ssh root@%s 'systemctl restart kafka'\"" % node, shell=True, check=True)

        return True

    def execute(self, skip_files_check=False):
        try:
            # Step 1
            if self.__get_confirmation("Please confirm restarting '%s': " % OUtils.Kafka.get_nodes()) is False:
                print("Rolling restart aborted")
                return

            # Step 2
            print("Verifying clean cluster status before restart ...")

            if (skip_files_check is False) and (self.__verify_cluster_clean_state() is True):
                print("\tOK")
            else:
                print("\tSkipped")

            # Step 3
            print("Executing rolling restart ...")

            for node in OUtils.Kafka.get_nodes():

                # Step 3a: restart service
                print("\tOn node %s: Restarting Kafka service ..." % node)
                if self.__execute_kafka_service_restart(node) is True:
                    print("\t\tOK")

                # Step 3b: check that cluster status returns to clean
                print("\tOn node %s: Checking under-replication ..." % node)

                for attempt in range(10):
                    time.sleep(5)

                    try:
                        under_replicated_replicas = OUtils.Kafka.metrics.get_total_under_replicated_replicas()

                    except ConnectionError:
                        print("\t\tFailed to connect. Will re-try.")
                        continue

                    if under_replicated_replicas == 0:
                        print("\t\tOK: No under-replicated replicas found: system stabilized back.\n")
                        break

                    print("\t\tFound %d under-replicated partitions." % under_replicated_replicas)
                    continue

                if under_replicated_replicas != 0:
                    raise Exception("System did not stabilize back after several attempts")

            print("Rolling restart completed.")
            return True

        except Exception as err:
            print("An exception occurred and the rolling restart was interrupted. Please verify cluster status.")
            print(str(err))
            return False
